/**
 *
 * $Id: adapnt.cpp,v 1.8 2005/04/03 08:27:47 mviara Exp $
 * $Author: mviara $
 * $Name:  $
 *
 *
 * $Log: adapnt.cpp,v $
 * Revision 1.8  2005/04/03 08:27:47  mviara
 * Added support for copy paste text.
 *
 * Revision 1.7  2005/04/02 07:26:29  mviara
 * Added partial support for copy/paste text.
 *
 * Revision 1.6  2005/03/28 12:53:09  mviara
 * Added support for xterm, fixed bug in attributr at begin of line.
 *
 * Revision 1.5  2005/01/24 17:31:21  mviara
 * Fixed bug in handling ESC[m
 *
 * Revision 1.4  2004/12/06 23:26:00  mviara
 * *** empty log message ***
 *
 * Revision 1.3  2004/12/06 22:12:13  mviara
 * Removed unused code.
 *
 * Revision 1.2  2004/12/06 07:31:05  mviara
 * Improved performance.
 *
 * Revision 1.1.1.1  2004/07/10 12:33:21  mviara
 * Migrated from RCS
 *
 * Win32 console driver support italian and english keyboard layout.
 * 
 */
#include "ansi.hpp"

static char rcsinfo[]= "$Id: adapnt.cpp,v 1.8 2005/04/03 08:27:47 mviara Exp $";

void CAdapterNT::SetColor(unsigned col)
{
	if (col != curColor)
	{
		curColor = col;
		SetAttribute();
	}
}


void CAdapterNT::SetAttribute()
{
	static WORD forColor[] =
			 {0,
			 FOREGROUND_BLUE,
			 FOREGROUND_GREEN,
			 FOREGROUND_BLUE|FOREGROUND_GREEN,
			 FOREGROUND_RED,
			 FOREGROUND_RED|FOREGROUND_GREEN,
			 FOREGROUND_RED|FOREGROUND_BLUE,
			 FOREGROUND_RED|FOREGROUND_BLUE|FOREGROUND_GREEN
				 };
	static WORD backColor[] =
			 {0,
			 BACKGROUND_BLUE,
			 BACKGROUND_GREEN,
			 BACKGROUND_BLUE|BACKGROUND_GREEN,
			 BACKGROUND_RED,
			 BACKGROUND_RED|BACKGROUND_GREEN,
			 BACKGROUND_RED|BACKGROUND_BLUE,
			 BACKGROUND_RED|BACKGROUND_BLUE|BACKGROUND_GREEN
				 };
	WORD attribute;
	
	attribute = forColor[curColor & 7];
	attribute |= backColor[(curColor >> 4) & 7];
	if (curColor & 8)
		attribute |= FOREGROUND_INTENSITY;
		
	SetConsoleTextAttribute(hStdOut,attribute);
}

char * CAdapterNT::Name()
{
	static char buffer[256];
	char * keyName;
	

	switch (kbdType)
	{
		case	1:
			keyName = "American";
			break;
		case	39:
			keyName = "Italian";
			break;
		default:
			keyName = "Unknown";
			break;
				
	}
	
	sprintf(buffer,"WIN32 Console $Revision: 1.8 $ Keyboard %s (%d)",keyName,padnum);

	return buffer;
}


CAdapterNT::~CAdapterNT()
{
	SetConsoleMode(hStdOut,oldConsoleOutMode);	
	SetConsoleMode(hStdIn,oldConsoleInMode);
	ShowCursor();
}

CAdapterNT::CAdapterNT(int k,int p)
{
	CONSOLE_SCREEN_BUFFER_INFO csb;
	DWORD newMode;

	padnum = p;
	
	//
	// Se la tastiera non e' specificata prende quella dal registry
	//
	if (k == 0)
	{
			char keyLayout[KL_NAMELENGTH];
			unsigned long  keyType;
			GetKeyboardLayoutName(keyLayout);
			sscanf(keyLayout,"%lx",&keyType);

			switch (keyType & 0xff)
			{
				case	LANG_ENGLISH:
					k = 1;
					break;
				case	LANG_ITALIAN:
					k = 39;
					
				default:
					k = 39;
					break;
			}
	}
	
	kbdType = k;
	hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	hStdIn  = GetStdHandle(STD_INPUT_HANDLE);
	GetConsoleCursorInfo(hStdOut,&cursorInfo);

	GetConsoleScreenBufferInfo(hStdOut,&csb);
	numRows = csb.dwSize.Y;
	numCols = csb.dwSize.X;
	key = 0;
	
	GetConsoleMode(hStdOut,&oldConsoleOutMode);
	newMode = oldConsoleOutMode;
	//newMode &= ~ENABLE_PROCESSED_OUTPUT;
	newMode &= ~ENABLE_WRAP_AT_EOL_OUTPUT;
	SetConsoleMode(hStdOut,newMode);
	
	GetConsoleMode(hStdIn,&oldConsoleInMode);
	newMode = oldConsoleInMode;
	newMode &= ~ENABLE_ECHO_INPUT;
	newMode &= ~ENABLE_MOUSE_INPUT;
	newMode &= ~ENABLE_PROCESSED_INPUT;
	SetConsoleMode(hStdIn,newMode);
	ShowCursor();
}

void CAdapterNT::Locate(int pos)
{
	COORD coord;

	
	coord.X = (short)(pos % Cols());
	coord.Y = (short)(pos / Cols());
	
	SetConsoleCursorPosition(hStdOut,coord);
}

void CAdapterNT::Print(int pos,char  ch,int num)
{
#if 1
	DWORD written;
	char buffer[80];
	
	memset(buffer,ch,sizeof(buffer));
	
	while (num)
	{
		Locate(pos);
		int col = pos % numCols;
		int n = numCols - col;
		if (n > num)
			n = num;
		WriteConsole(hStdOut,buffer,n,&written,NULL);
		pos += n;
		num -= n;
	}
#else
	int i;
	
	assert (num >= 0);
	assert (pos < Size());
	
	//SetAttribute();

	if (bufferOutLen)
	{
		if (bufferOutLen+bufferOutPos != pos)
		{
			FlushBuffer();
			bufferOutPos = pos;
		}
	}
	else
		bufferOutPos = pos;
		

	
	if (pos + num >= Size())
		num = Size() - pos - 1;
	

	assert (num >= 0);
	
	while (num)
	{
		if (bufferOutLen == sizeof(bufferOut))
		{
			i = bufferOutPos + bufferOutLen;
			FlushBuffer();
			bufferOutPos = i;
		}
		
		i = sizeof(bufferOut) - bufferOutLen;
		if (num < i )
			i = num;
		
		memset(bufferOut+bufferOutLen,ch,i);
		bufferOutLen += i;
		num -= i;
	}
#endif
}


void CAdapterNT::PrintBuffer(int pos,char *buffer,int num)
{
#if 1
	DWORD written;


	while (num)
	{
		Locate(pos);
		int col = pos % numCols;
		int n = numCols - col;
		if (n > num)
			n = num;
		WriteConsole(hStdOut,buffer,n,&written,NULL);
		pos += n;
		num -= n;
		buffer += n;
	}

#else
	int i;
	//SetAttribute();
	
	if (pos + num >= Size())
		num = Size() - pos - 1;

	assert (num >= 0);

	if (bufferOutLen)
	{
		if (bufferOutLen+bufferOutPos != pos)
		{
			FlushBuffer();
			bufferOutPos = pos;
		}
	}
	else
		bufferOutPos = pos;
		

	
	
	while (num)
	{
		if (bufferOutLen == sizeof(bufferOut))
		{
			i = bufferOutPos + bufferOutLen;
			FlushBuffer();
			bufferOutPos = i;
		}
		
		i = sizeof(bufferOut) - bufferOutLen;
		if (num < i )
			i = num;
		
		memcpy(bufferOut+bufferOutLen,buffer,i);
		bufferOutLen += i;
		num -= i;
		buffer += i;
	}
#endif
}


static void KeyEvent(int kbdType,KEY_EVENT_RECORD *k,unsigned& key,int padnum,KSTATE & kstate)
{
	static DWORD oldState = 0;
	WORD vk;
	
	DWORD state;
	//DWORD shift,control,caps,alt,num,numshift=0,altgr;
	int numshift = 0,sh;
	static char *shiftNumUsa = ")!@#$%^&*(";
	static char *shiftNumIta = "=!\"$%&/()";
	char dot = '.';

	if (kbdType == 39)
		dot = ',';
	
	state = k->dwControlKeyState;

	//
	// Qui gestisce gli scroll lock
	//
	if ((state & SCROLLLOCK_ON) != (oldState & SCROLLLOCK_ON))
	{
		oldState = state;
		if (state & SCROLLLOCK_ON)
		{
			key = CAnsi::KEYSCROLLOFF;
		}
		else
		{
			key = CAnsi::KEYSCROLLON;
		}
		return;
	}

	
	
	if (k->bKeyDown == FALSE)
		return;
	
	vk = k->wVirtualKeyCode;
	
	kstate.flag.shift   = (state & (SHIFT_PRESSED)) != 0;
	kstate.flag.ctrl    = (state & (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED)) != 0;
	kstate.flag.caps    = (state & CAPSLOCK_ON) != 0;
	kstate.flag.alt	    = (state & (RIGHT_ALT_PRESSED|LEFT_ALT_PRESSED)) != 0;
	kstate.flag.nums    = (state & NUMLOCK_ON) != 0;
	kstate.flag.altgr   = (state & RIGHT_ALT_PRESSED) != 0;

	if (kbdType == 1)
		kstate.flag.altgr = 0;
	
	if (kstate.flag.nums)
		numshift = !kstate.flag.shift;
	
	
	//AnsiPrintf("VK = %d %x, State %x, Altgr %d\n",vk,vk,state,altgr);
	
	switch (vk)
	{
		case	0xe2:
			if (kstate.flag.shift)
				key = '>';
			else
				key = '<';
			break;
			
		case	0xc0:
			
			if (kstate.flag.altgr)
			{
				key = '@';
				break;
			}
			
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = '~';
				else
					key = '`';
			}
			else
			{
				if (kstate.flag.shift)
					key = '';
				else
				        key = '';
			}
			
			break;
		case	0xbd:
			if (kstate.flag.shift)
				key = '_';
			else
				key = '-';
			break;
		case	0xbb:
			if (kstate.flag.altgr)
			{
				key = ']';
				break;
			}
			
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = '+';
				else
					key = '=';
			}
			else
			{
				if (kstate.flag.shift)
					key = '*';
				else
					key = '+';
			}
			break;
		case	0xdc:
			if (kstate.flag.shift)
				key = '|';
			else
				key = '\\';
			break;
			
		case	0xdb:
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = '{';
				else
					key = '[';
			}
			else
			{
				if (kstate.flag.shift)
					key = '?';
				else
					key = '\'';
			}
			
			break;
		
		case	0xdd:
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = '}';
				else
					key = ']';
			}
			else
			{
				if (kstate.flag.shift)
					key = '^';
				else
					key = '';
			}
			break;
		
		case	0xba:
			if (kstate.flag.altgr)
			{
				key = '[';
				break;
			}
			
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = ':';
				else	
					key = ';';
				break;
			}
			else
			{
				if (kstate.flag.shift)
					key = '';
				else
					key = '';
			}
			break;
			
		case	0xde:
			if (kstate.flag.altgr)
			{
				key = '#';
				break;
			}
			
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = '"';
				else
					key = '\'';
			}
			else
			{
				if (kstate.flag.shift)
					key = '';
				else
					key = '';
			}
			
			break;
		
		case	0xbe:
			if (kstate.flag.shift)
			{
				if (kbdType == 1)
					key = '>';
				else
					key = ':';
			}
			else
				key = '.';
			break;
			
		case	0xbc:
			if (kstate.flag.shift)
			{
				if (kbdType == 1)
					key = '<';
				else
					key = ';';
			}
			else
				key = ',';
			break;
			
		case	0xbf:
			if (kbdType == 1)
			{
				if (kstate.flag.shift)
					key = '?';
				else
					key = '/';
			}
			else
			{
				if (kstate.flag.shift)
					key = '';
				else
					key = '';
			}
			break;
			
		case	VK_BACK:
			key = 8;
			break;
			
		case	VK_TAB:
			key = 9;
			break;
			
		case	VK_ESCAPE:
			key = 27;
			break;
			
		case	VK_RETURN:
			key = 13;
			break;
			
		case	VK_SHIFT:
		case	VK_CONTROL:
		case	VK_MENU:
		case	VK_CAPITAL:
			break;
			//
			// Lettere dalla  A alla Z
			//
		case	'A':
		case	'B':
		case	'C':
		case	'D':
		case	'E':
		case	'F':
		case	'G':
		case	'H':
		case	'I':
		case	'J':
		case	'K':
		case	'L':
		case	'M':
		case	'N':
		case	'O':
		case	'P':
		case	'Q':
		case	'R':
		case	'S':
		case	'T':
		case	'U':
		case	'V':
		case	'W':
		case	'X':
		case	'Y':
		case	'Z':
			if (kstate.flag.alt)
			{
				key = k->wVirtualScanCode << 8;
				if (kstate.flag.ctrl)
					key |= 1;
				break;
			}
			else if (kstate.flag.ctrl)
			{
				key =  vk - '@';
				break;
			}

			sh = kstate.flag.shift;
			
			if (kstate.flag.caps)	// scambia shift se caps
				sh =!sh;
			
			
			if (sh)
				key = vk;
			else
				key =  vk + 0x20;
			break;
			
		case	'0':
		case	'1':
		case	'2':
		case	'3':
		case	'4':
		case	'5':
		case	'6':
		case	'7':
		case	'8':
		case	'9':
			if (kstate.flag.shift)
			{
				if (kbdType == 1)
					key = shiftNumUsa[vk - '0'];
				else
					key = shiftNumIta[vk - '0'];
			}
			else
				key = vk;
			break;
		case	VK_F11:
				key = 0x8500;
				break;
		case	VK_F12:
				key = 0x8600;
				break;
		case	VK_F1:
		case	VK_F2:
		case	VK_F3:
		case	VK_F4:
		case	VK_F5:
		case	VK_F6:
		case	VK_F7:
		case	VK_F8:
		case	VK_F9:
		case	VK_F10:
				
				if	(kstate.flag.alt)
					key = (k->wVirtualScanCode+0x2d) << 8;
				else if (kstate.flag.shift)
				{
					key = (k->wVirtualScanCode + 0x19) << 8;
				}
				else
					key =  k->wVirtualScanCode << 8;
			break;

			
		case	VK_MULTIPLY:
			key  = '*';
			break;

		case	VK_ADD:
			key = '+';
			break;

		case	VK_DIVIDE:
			key = '/';
			break;
		case	VK_SUBTRACT:
			key = '-';
			break;

		case	VK_DECIMAL:
			if (numshift)
			{
				if (kbdType == 1)
					key = dot;
				else
					key = dot;
				break;
			}
			key =  k->wVirtualScanCode << 8;
			break;

		case	VK_SPACE:
			key = ' ';
			break;

		case	VK_NUMPAD0:
		case	VK_NUMPAD1:
		case	VK_NUMPAD2:
		case	VK_NUMPAD3:
		case	VK_NUMPAD4:
		case	VK_NUMPAD5:
		case	VK_NUMPAD6:
		case	VK_NUMPAD7:
		case	VK_NUMPAD8:
		case	VK_NUMPAD9:
			key = vk - VK_NUMPAD0 + '0';
			break;
		case	VK_DELETE:
			if (padnum)
			{
				key = dot;
				break;
			}
				
		case	VK_PRIOR:
		case	VK_NEXT:
		case	VK_END:
		case	VK_HOME:
		case	VK_RIGHT:
		case	VK_LEFT:
		case	VK_UP:
		case	VK_DOWN:
		case	VK_INSERT:
			key =  k->wVirtualScanCode << 8;
			break;
	}

}

unsigned CAdapterNT::Getkey(int kbdtype,int padnum)
{
	INPUT_RECORD record;
	DWORD nevent=0;
	unsigned key;
	
	nevent = 0;
	PeekConsoleInput(hStdIn,&record,1,&nevent);
	if (nevent == 0)
		return 0;
	
	nevent = 0;
	
	ReadConsoleInput(hStdIn,&record,1,&nevent);
	if (nevent == 0)
		return 0;

	key = 0;
		
	switch (record.EventType)
	{
		case	KEY_EVENT:
			KeyEvent(kbdtype,&record.Event.KeyEvent,key,padnum,kstate);
	}


	return key;
}


int CAdapterNT::Kbhit()
{
	
	if (key)
		return 1;
	
	key = Getkey(kbdType,padnum);

	if (key)
		return 1;
	
	return 0;
}

unsigned CAdapterNT::Getch()
{
	unsigned oldKey;
	
	while (!Kbhit())
		;

	oldKey = key;
	key = 0;
	
	return oldKey;
	
}


void CAdapterNT::HideCursor()
{
	cursorInfo.bVisible = FALSE;
	SetConsoleCursorInfo(hStdOut,&cursorInfo);
}


void CAdapterNT::ShowCursor()
{
	cursorInfo.bVisible = TRUE;
	SetConsoleCursorInfo(hStdOut,&cursorInfo);
}

void CAdapterNT::BigCursor()
{
	int oldSize = cursorInfo.dwSize;
	BOOLEAN v = cursorInfo.bVisible;
	cursorInfo.dwSize   = 100;
	cursorInfo.bVisible = TRUE;
	SetConsoleCursorInfo(hStdOut,&cursorInfo);
	cursorInfo.dwSize = oldSize;
	cursorInfo.bVisible = v;
}

void CAdapterNT::LittleCursor()
{
	SetConsoleCursorInfo(hStdOut,&cursorInfo);
}
